SystemOrganization addCategory: #'Xpf-Multipoint-Core'!

!MenuMorph methodsFor: '*xpf-multipoint-core-override' stamp: 'mig 6/20/2009 19:06'!
popUpAt: aPoint forHand: hand in: aWorld allowKeyboard: aBoolean 
	"Present this menu at the given point under control of the given  
	hand."
	| evt |
 	MultiPointerManager multiPointerActive ifFalse: [
 		aWorld submorphs
 			select: [ :each | (each isKindOf: MenuMorph)
 				and: [each stayUp not]]
 			thenDo: [ :menu | menu delete]].

	self items isEmpty
		ifTrue: [^ self].

	MenuIcons decorateMenu: self.

	(self submorphs
		select: [:m | m isKindOf: UpdatingMenuItemMorph])
		do: [:m | m updateContents].
	"precompute width"
	self
		positionAt: aPoint
		relativeTo: (selectedItem
				ifNil: [self items first])
		inWorld: aWorld.
	aWorld addMorphFront: self.
	"Acquire focus for valid pop up behavior"
	hand newMouseFocus: self.
	aBoolean
		ifTrue: [hand newKeyboardFocus: self].
	evt := hand lastEvent.
	(evt isKeyboard
			or: [evt isMouse
					and: [evt anyButtonPressed not]])
		ifTrue: ["Select first item if button not down"
			self moveSelectionDown: 1 event: evt].
	self updateColor.
	self changed! !

!HandMorph methodsFor: '*xpf-multipoint-core' stamp: 'SR 5/31/2009 14:40'!
isXpfMultiPointerHandMorph

	^ false! !

!HandMorph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 8/18/2009 16:58'!
processEvents
	"Process user input events from the local input devices."
	| evtBuf hadAny shouldProcessEvents |

	shouldProcessEvents := [MultiPointerManager moving not] ifError: [true].
	
	shouldProcessEvents ifTrue: [
		ActiveEvent ifNotNil: [
			"Meaning that we were invoked from within an event response. Make sure z-order is up to date"
			self mouseOverHandler processMouseOver: lastMouseEvent]].
	
	hadAny := false.
	
	[(evtBuf := Sensor nextEvent) notNil]
		whileTrue: [ | evt type |
			type := evtBuf first.
			evt := nil.
			
			shouldProcessEvents 
				ifTrue: [
					"normal behaviour"
					type = EventTypeMouse ifTrue: [evt := self generateMouseEvent: evtBuf].
					type = EventTypeKeyboard ifTrue: [evt := self generateKeyboardEvent: evtBuf].
					type = EventTypeDragDropFiles ifTrue: [evt := self generateDropFilesEvent: evtBuf].
				
					evt ifNotNil: [self handleEvent: evt. hadAny := true]]
				ifFalse: [
					"multipointer behaviour: process keyboard only and forward to another hand"
					"type = EventTypeKeyboard ifTrue: [evt := self generateKeyboardEvent: evtBuf]."
					
					"sometimes even generateKeyboardEvent returns mouse events (for mouse wheels)"
					"(evt notNil and: [evt isKeyboard]) ifTrue: [
							MultiPointerManager instance keyboardHand enqueueEvent: evt]"]].
	
	
	shouldProcessEvents ifTrue: [
		"note: if we come here we didn't have any mouse events"
		mouseClickState ifNotNil: [
			"No mouse events during this cycle. Make sure click states time out accordingly"
			mouseClickState handleEvent: lastMouseEvent asMouseMove from: self].
	
		hadAny ifFalse: ["No pending events. Make sure z-order is up to date"
			self mouseOverHandler processMouseOver: lastMouseEvent]].
! !

HandMorph subclass: #MultiPointerHandMorph
	instanceVariableNames: 'queue lastPosition buttons pointer normalCursor handColor movedTimestamp'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

!MultiPointerHandMorph class methodsFor: 'constants' stamp: 'mig 6/4/2009 17:27'!
cursorSize

	^ 35! !

!MultiPointerHandMorph class methodsFor: 'constants' stamp: 'SR 5/31/2009 14:35'!
displayTimeout

	^ 10000! !

!MultiPointerHandMorph class methodsFor: 'icons' stamp: 'SR 5/25/2009 15:51'!
keyboardIcon

	^ Form fromBinaryStream: (Base64MimeConverter mimeDecodeToBytes: self keyboardIconContents readStream)! !

!MultiPointerHandMorph class methodsFor: 'icons' stamp: 'SR 5/25/2009 17:29'!
keyboardIconContents

	^ 'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAGXRFWHRTb2Z0d2FyZQBBZG9iZSBJ
bWFnZVJlYWR5ccllPAAAAdxJREFUeNrEU0tLG1EU/iadgggVMUJomhQxPiuU+CjURTZWEJFuuhLd
uelfCP6AbsWlP6BQcVtaqC/sxkWhDUUIjSUSDAaNGJLMZJI799Fzb2KEFgriwgPffOfemXu+c8+Z
YymlcBcL4I5mXzuZTOat7/trhAZZrV6vVz3PKxJStVrtgrhMfEzYTiaT/J8A9ME6HQQhmUgkNm6d
geM4D0n1MwXavNUVvp6IU+KI63qgAAvdz+ZKO78a4EISBHwuwH0BRuz7es0Ns6af1xlEQnYVGJ0y
EZVyoBsjNeihuySIpdR7D8i3yLeJgYP0VaTdhYHwI+QvXAw+6ULuvIqRaBeyZxWM9XXjOF9GfKAH
6VwJL4Z7zZ7OkDEOa/83UyHbQb7oGkX9QitygjDX0FBtjseCmBgKYu9HAbs/iwhwLs3BmfHHmJ0M
010lFl5G0WACbxJ9qDOJpVcxWkuszA+Zw9p0QF0P69ORq8IdHk4KFXCulW9UdQZ+mykzcZPZ8/4g
tlMF2KbKpPp6+ik+7GWxONP/37a9e5/C6nIcHw9PjZC19e1KRTsZRKvqUkpT4TYr1dpXxteTo2ST
v3w/g91g/DLtqt6/+tv0W/+AyVKIZoEJktAawUvr3qfxjwADAPOBecF+sUNmAAAAAElFTkSuQmCC'! !

!MultiPointerHandMorph class methodsFor: 'misc' stamp: 'mig 5/27/2009 12:14'!
scaledCursor: size colored: aColor

	| form canvas leftPoint rightPoint barOffset borderWidth baseOffset |
	form := Form extent: size@size depth: 32.
	canvas := FormCanvas on: form.
	borderWidth := ((size / 20) max: 1) rounded.
	baseOffset := borderWidth@borderWidth.
	
	rightPoint := Point r: size*0.75 degrees: 45.
	leftPoint := Point r: size*0.75 degrees: 90.
	barOffset := Point r: size*0.2 degrees: 62.5.
	
	canvas
		drawPolygon: {
			baseOffset.
			baseOffset + rightPoint.
			baseOffset + (leftPoint * 0.33) + (rightPoint * 0.67).
			baseOffset + (leftPoint * 0.33) + (rightPoint * 0.67) + barOffset.
			baseOffset + (leftPoint * 0.67) + (rightPoint * 0.33) + barOffset.
			baseOffset + (leftPoint * 0.67) + (rightPoint * 0.33).
			baseOffset + leftPoint} 
		fillStyle: aColor
		borderWidth: borderWidth
		borderColor: Color black.
	
	^ (form asCursorForm) offset: baseOffset negated; yourself! !

!MultiPointerHandMorph class methodsFor: 'constants' stamp: 'mig 6/4/2009 17:39'!
waveSteps

	^ (1.0 to: 3.0 by: 0.3), (3.0 to: 1.0 by: -0.5)! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/29/2009 17:49'!
buttons

	^ buttons! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/6/2009 22:21'!
buttons: newButtons

	buttons := newButtons.! !

!MultiPointerHandMorph methodsFor: 'updating' stamp: 'mig 5/19/2009 20:53'!
changed

	bounds := bounds origin extent: self currentCursor extent.
	super changed.! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 20:04'!
createMorphicButtonEvents: aState

	| result |
	result := OrderedCollection new.
	
	0 to: 2 do: [:each | | buttonBit |
		buttonBit := 1 bitShift: (2-each).
	
		(aState buttons anyMask: (1 bitShift: each*2)) ifTrue: [
			self buttons: (self buttons bitOr: buttonBit).
			result add: (self pvtMouseButtonEvent: #mouseDown with: buttonBit)].
	
		(aState buttons anyMask: (1 bitShift: (each*2)+1)) ifTrue: [
			self buttons: (self buttons bitAnd: (buttonBit bitInvert)).
			result add: (self pvtMouseButtonEvent: #mouseUp with: buttonBit)]].
	
	^ result! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 6/24/2009 18:59'!
createMorphicMoveEvent: aState

	| newPosition evt |
	newPosition := (self lastPosition + aState delta) adhereTo: Display fullBoundingBox.
	
	evt := MouseMoveEvent new
		setType: #mouseMove
		startPoint: self lastPosition
		endPoint: newPosition
		trail: nil
		buttons: self buttons
		hand: self
		stamp: Time millisecondClockValue.
	
	(self lastPosition = newPosition) ifFalse: [
		self resetTimeout].
	self lastPosition: newPosition.

	^ evt
! !

!MultiPointerHandMorph methodsFor: 'cursor' stamp: 'mig 5/18/2009 14:18'!
currentCursor

	^ {self temporaryCursor. self normalCursor. NormalCursor} detect: [:cursor| cursor notNil]! !

!MultiPointerHandMorph methodsFor: 'drawing' stamp: 'SR 8/18/2009 15:34'!
drawOn: aCanvas 
	"Draw the hand itself (i.e., the cursor)."
	
	self hasMovedInLastTime ifFalse: [^ self].
	
	aCanvas 
		paintImage: self currentCursor
		at: bounds topLeft.

	"(self isKeyboardHand and: [self temporaryCursor isNil]) ifTrue: [
		aCanvas
			translucentImage: self class keyboardIcon
			at: (bounds bottomRight - self class keyboardIcon extent)]."! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'mig 6/8/2009 16:24'!
enqueueEvent: anEvent

	"use modifier keys from keyboard events (16r78 is the modifier key mask)"
	(anEvent notNil and: [anEvent isKeyboard]) ifTrue: [
		self buttons: ((self buttons bitAnd: 16r78 bitInvert) bitOr: (anEvent buttons bitAnd: 16r78))].

	self queue nextPut: anEvent.
! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'mig 6/24/2009 19:03'!
enqueueEvents: anEventCollection

	anEventCollection do: [:event |
		self enqueueEvent: event].! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 16:15'!
handColor

	^ handColor! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 6/4/2009 17:28'!
handColor: aColor

	handColor := aColor.
	self normalCursor: (self class scaledCursor: self class cursorSize colored: aColor).! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:28'!
handId

	^ self pointer pointerID asString
! !

!MultiPointerHandMorph methodsFor: 'events-processing' stamp: 'mig 6/3/2009 11:06'!
handleEvent: anEvent
	
	(anEvent isMouse and: [anEvent isMouseDown]) ifTrue: [
			(anEvent redButtonPressed and: [anEvent yellowButtonPressed]) ifTrue: [	self wave]].
		
	^ super handleEvent: anEvent! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/27/2009 10:56'!
hasMovedInLastTime

	^ self movedTimestamp notNil 
		and: [Time millisecondClockValue - self movedTimestamp < self class displayTimeout]! !

!MultiPointerHandMorph methodsFor: 'initialization' stamp: 'mig 5/27/2009 10:26'!
initialize

	super initialize.
	self
		queue: SharedQueue new;
		lastPosition: 0@0;
		buttons: 0.
! !

!MultiPointerHandMorph methodsFor: 'testing' stamp: 'SR 5/31/2009 14:39'!
isXpfMultiPointerHandMorph

	^ true! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 13:38'!
lastPosition

	^ lastPosition! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 20:06'!
lastPosition: anObject

	lastPosition := anObject.! !

!MultiPointerHandMorph methodsFor: 'updating' stamp: 'mig 6/3/2009 10:07'!
moveToPosition: aPosition
	"Issue a mouse move event to make the receiver appear at the given position"
	self lastPosition: aPosition.
	self handleEvent: (MouseMoveEvent new
		setType: #mouseMove 
		startPoint: self position 
		endPoint: aPosition 
		trail: (Array with: self position with: aPosition)
		buttons: 0
		hand: self
		stamp:  Time millisecondClockValue)! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/27/2009 10:25'!
movedTimestamp

	^ movedTimestamp! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/27/2009 10:25'!
movedTimestamp: anObject

	movedTimestamp := anObject.! !

!MultiPointerHandMorph methodsFor: 'drawing' stamp: 'SR 4/24/2009 13:39'!
needsToBeDrawn

	^ true! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/6/2009 10:41'!
normalCursor

	^ normalCursor! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 20:53'!
normalCursor: anObject

	normalCursor := anObject.
	self changed.! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:22'!
pointer

	^ pointer! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:22'!
pointer: anObject

	pointer := anObject.! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'mig 5/7/2009 13:54'!
processEvents
	
	ActiveEvent ifNotNil: [
		"Meaning that we were invoked from within an event response. Make sure z-order is up to date"
		self mouseOverHandler processMouseOver: lastMouseEvent].
	
	self queue isEmpty ifTrue: [
		"No pending events. Make sure z-order is up to date"
		self mouseOverHandler processMouseOver: lastMouseEvent.
		^ self].

	[self queue isEmpty] whileFalse: [
		self handleEvent: self queue next].! !

!MultiPointerHandMorph methodsFor: 'private' stamp: 'mig 5/19/2009 20:04'!
pvtMouseButtonEvent: aSymbol with: anInteger

	^ MouseButtonEvent new 
		setType: aSymbol
		position: self lastPosition
		which: anInteger
		buttons: self buttons
		hand: self
		stamp: Time millisecondClockValue! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'SR 4/24/2009 13:38'!
queue

	^ queue! !

!MultiPointerHandMorph methodsFor: 'accessing' stamp: 'mig 5/6/2009 22:21'!
queue: anObject

	queue := anObject.! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'mig 6/24/2009 19:04'!
readEvent

	| state |
	state := self pointer state.
	
	state delta isZero ifFalse: [
		self enqueueEvent: (self createMorphicMoveEvent: state)].
	state buttons isZero ifFalse: [
		self enqueueEvents: (self createMorphicButtonEvents: state)].! !

!MultiPointerHandMorph methodsFor: 'event handling' stamp: 'mig 6/5/2009 11:05'!
resetTimeout

	| hasMovedBefore |
	
	hasMovedBefore := self hasMovedInLastTime.
	
	self movedTimestamp: Time millisecondClockValue.
	
	(hasMovedBefore not and: [self hasMovedInLastTime]) ifTrue: [self wave].! !

!MultiPointerHandMorph methodsFor: 'drawing' stamp: 'mig 6/4/2009 17:37'!
wave

	[
		self class waveSteps do: [:factor |
			self showTemporaryCursor: 
				(self class 
					scaledCursor: self class cursorSize * factor 
					colored: self handColor).
			(Delay forMilliseconds: 20) wait].
		
		self showTemporaryCursor: nil.
	] fork.! !

Object subclass: #MultiPointer
	instanceVariableNames: 'pointerID'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

MultiPointer class
	instanceVariableNames: 'bufferSemaphore'!

!MultiPointer commentStamp: 'mig 5/19/2009 19:49' prior: 0!
This class provides abstractions to the MultiPointer plugin. Instances are bound to a specific pointer in the operating system and provide messages for retrieving information about them.!

MultiPointer class
	instanceVariableNames: 'bufferSemaphore'!

!MultiPointer class methodsFor: 'instance creation' stamp: 'SR 8/19/2009 13:57'!
allPointers

	^ {}
	"^ (0 to: self getPointerCount - 1) collect: [:id | self newForID: id]"! !

!MultiPointer class methodsFor: 'accessing' stamp: 'SR 8/13/2009 14:28'!
bufferSemaphore

	^ bufferSemaphore ifNil: [self bufferSemaphore: Semaphore new. bufferSemaphore]! !

!MultiPointer class methodsFor: 'accessing' stamp: 'SR 8/13/2009 14:11'!
bufferSemaphore: anObject

	bufferSemaphore := anObject.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/19/2009 13:52'!
eventAvailable

	^ self pvtEventAvailable ~= 0! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:21'!
getDeviceCount

	<primitive: 'primitiveMultiPointerDeviceCount' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:29'!
getDeviceLively: aPointerID

	^ (self pvtGetDeviceLively: aPointerID) ~= 0! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:57'!
getNextEvent: array

	<primitive: 'primitiveMultiPointerGetNextEvent' module: 'MultiPointerPlugin'>
	"self primitiveFailed."

! !

!MultiPointer class methodsFor: 'instance creation' stamp: 'mig 5/19/2009 10:51'!
newForID: aPointerID

	^ self new pointerID: aPointerID! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/19/2009 13:52'!
pvtEventAvailable

	<primitive: 'primitiveMultiPointerEventAvailable' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:18'!
pvtGetDeviceLively: aPointerID

	<primitive: 'primitiveMultiPointerDeviceLively' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'system startup' stamp: 'SR 8/13/2009 14:20'!
registerSemaphores

	self setBufferSemaphoreHandle: (Smalltalk registerExternalObject: self bufferSemaphore).! !

!MultiPointer class methodsFor: 'accessing' stamp: 'mig 5/19/2009 14:58'!
resetAll

	self allPointers do: [:pointer | pointer isLively ifTrue: [pointer reset]].! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:19'!
setBufferSemaphoreHandle: aSemaIndex

	<primitive: 'primitiveMultiPointerSetBufferSemaphore' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'system startup' stamp: 'SR 8/13/2009 14:21'!
unregisterSemaphores

	self setBufferSemaphoreHandle: 0.
	Smalltalk unregisterExternalObject: self bufferSemaphore.
	self bufferSemaphore: nil.
! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:20'!
useMultiInput

	<primitive: 'primitiveMultiPointerUseMultiInput' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer class methodsFor: 'primitives' stamp: 'SR 8/13/2009 14:20'!
useOSInput

	<primitive: 'primitiveMultiPointerUseOSInput' module: 'MultiPointerPlugin'>
	self primitiveFailed.! !

!MultiPointer methodsFor: 'comparing' stamp: 'mig 5/19/2009 10:58'!
= aPointer

	^ self pointerID = aPointer pointerID! !

!MultiPointer methodsFor: 'testing' stamp: 'mig 5/19/2009 10:13'!
isLively

	^ self class getPointerLively: self pointerID! !

!MultiPointer methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:12'!
pointerID

	^ pointerID! !

!MultiPointer methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:19'!
pointerID: anInteger

	pointerID := anInteger.! !

!MultiPointer methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:48'!
reset

	self class getPointerCurrentState: self pointerID.! !

!MultiPointer methodsFor: 'accessing' stamp: 'mig 5/19/2009 10:18'!
state

	^ self class getPointerCurrentState: self pointerID! !

SharedPool subclass: #MultiPointerConstants
	instanceVariableNames: ''
	classVariableNames: 'EventTypeNone EventTypeMenu DeviceTypeMouse DeviceTypeKeyboard RedButtonBit ShiftKeyBit EventTypeDeviceDiscovered EventTypeDeviceRemoved BlueButtonBit EventTypeWindow EventTypeMouse EventKeyChar EventKeyDown YellowButtonBit OptionKeyBit EventTypeDragDropFiles CommandKeyBit EventTypeKeyboard EventKeyUp CtrlKeyBit'
	poolDictionaries: ''
	category: 'Xpf-Multipoint-Core'!

!MultiPointerConstants class methodsFor: 'pool initialization' stamp: 'SR 8/18/2009 16:20'!
initialize
	"EventSensorConstants initialize"
	RedButtonBit := 4.
	BlueButtonBit := 2.
	YellowButtonBit := 1.

	ShiftKeyBit := 1.
	CtrlKeyBit := 2.
	OptionKeyBit := 4.
	CommandKeyBit := 8.

	"Types of events"
	EventTypeNone := 0.
	EventTypeMouse := 1.
	EventTypeKeyboard := 2.
	EventTypeDragDropFiles := 3.
	EventTypeMenu := 4.
	EventTypeWindow := 5.
	EventTypeDeviceDiscovered := 6.
	EventTypeDeviceRemoved := 7.

	"Types of devices"
	DeviceTypeMouse := 1.
	DeviceTypeKeyboard := 2.

	"Press codes for keyboard events"
	EventKeyChar := 0.
	EventKeyDown := 1.
	EventKeyUp := 2.
! !

Object subclass: #MultiPointerManager
	instanceVariableNames: 'handMorphs colors queueProcessor moving'
	classVariableNames: ''
	poolDictionaries: 'MultiPointerConstants'
	category: 'Xpf-Multipoint-Core'!

MultiPointerManager class
	instanceVariableNames: 'instance'!

!MultiPointerManager commentStamp: 'mig 5/19/2009 19:47' prior: 0!
A MultiPointerManager controls all MultiPointer actions. There should be only one MultiPointerManager instance in a running system. Instances are created and deleted upon startup and shutdown.

colors: is a dictionary having colors as keys and usage counts as values.
deviceUpdateProcess and movementProcess: hold the two forked processes for MultiPointer.
handMorphs: is a list of all multipointer hands.
keyboardHand: is the hand which sends keyboard events into the world.
!

MultiPointerManager class
	instanceVariableNames: 'instance'!

!MultiPointerManager class methodsFor: 'class initialization' stamp: 'mig 5/19/2009 11:44'!
initialize

	super initialize.
	Smalltalk addToStartUpList: self.
	Smalltalk addToShutDownList: self.
	! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:20'!
instance

	^ instance! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:20'!
instance: anObject

	instance := anObject.! !

!MultiPointerManager class methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:37'!
multiPointerActive

	^ self instance notNil and: [self instance multiPointerActive]! !

!MultiPointerManager class methodsFor: 'testing' stamp: 'SR 5/6/2009 10:19'!
pluginAvailable

	^ SmalltalkImage current listBuiltinModules anySatisfy: [:each | each startsWith: 'MultiPointerPlugin']! !

!MultiPointerManager class methodsFor: 'system startup' stamp: 'mig 5/19/2009 11:55'!
shutDown: aboutToQuit 

	self instance ifNotNil: [
		self instance shutdown.
		self instance: nil].
	
	super shutDown: aboutToQuit.! !

!MultiPointerManager class methodsFor: 'event processing' stamp: 'mig 5/19/2009 11:57'!
startMoving

	self instance ifNotNil: [ self instance startMoving ]! !

!MultiPointerManager class methodsFor: 'system startup' stamp: 'mig 5/19/2009 11:54'!
startUp: aboutToQuit 

	self pluginAvailable ifTrue: [
		self instance: self new.
		self instance startup].
	
	super startUp: aboutToQuit.! !

!MultiPointerManager class methodsFor: 'event processing' stamp: 'mig 5/19/2009 11:57'!
stopMoving

	self instance ifNotNil: [ self instance stopMoving ]! !

!MultiPointerManager methodsFor: 'hands' stamp: 'SR 8/18/2009 15:32'!
addAllHands

	self handMorphs do: [:hand | 	
		World addHand: hand.
		hand changed].
! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 16:19'!
colors

	^ colors! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 16:19'!
colors: anObject

	colors := anObject.! !

!MultiPointerManager methodsFor: 'hands' stamp: 'mig 5/19/2009 20:33'!
createHandMorphFor: aPointer
	
	| hand |
	(self handMorphs anySatisfy: [:each | each pointer = aPointer]) 
		ifTrue: [^ self].

	hand := MultiPointerHandMorph new 
		pointer: aPointer;
		handColor: self getColor.
	self handMorphs add: hand.
	self multiPointerActive ifTrue: [World addHand: hand].
! !

!MultiPointerManager methodsFor: 'hands' stamp: 'SR 8/18/2009 15:32'!
destroyHandMorphFor: aPointer

	(self handMorphs detect: [:each | each pointer = aPointer] ifNone: [nil])
		ifNotNilDo: [:morph |
			self handMorphs remove: morph.
			World removeHand: morph.
			self releaseColor: morph handColor].
! !

!MultiPointerManager methodsFor: 'hand colors' stamp: 'mig 5/19/2009 16:26'!
getColor

	| color |
	color := self colors keyAtValue: self colors values min.
	self colors at: color put: (self colors at: color) + 1.
	^ color! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:21'!
handMorphs

	^ handMorphs! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'mig 5/19/2009 11:21'!
handMorphs: anObject

	handMorphs := anObject.! !

!MultiPointerManager methodsFor: 'event processing' stamp: 'SR 8/19/2009 14:11'!
handleDetectedEvent: evt

	self inform: 'Device found'.! !

!MultiPointerManager methodsFor: 'event processing' stamp: 'SR 8/18/2009 16:41'!
handleRemovedEvent: evt

	self inform: 'Device removed'! !

!MultiPointerManager methodsFor: 'initialize-release' stamp: 'SR 8/18/2009 15:38'!
initialize

	super initialize.
	
	self
		moving: false;
		handMorphs: OrderedCollection new;
		initializeColors.! !

!MultiPointerManager methodsFor: 'hand colors' stamp: 'mig 5/19/2009 16:44'!
initializeColors

	self colors: (Dictionary newFrom: {
		Color red -> 0.
		Color orange -> 0.
		Color green -> 0.
		Color blue -> 0.
	}).
! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/18/2009 15:38'!
moving

	^ moving! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/18/2009 15:38'!
moving: anObject

	moving := anObject! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/18/2009 15:31'!
multiPointerActive

	^ self queueProcessor notNil! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/18/2009 15:31'!
queueProcessor

	^ queueProcessor! !

!MultiPointerManager methodsFor: 'accessing' stamp: 'SR 8/18/2009 15:31'!
queueProcessor: anObject

	queueProcessor := anObject! !

!MultiPointerManager methodsFor: 'event processing' stamp: 'SR 8/19/2009 14:12'!
queueProcessorBlock

	^ [[
		MultiPointer bufferSemaphore wait.
		[MultiPointer eventAvailable] whileTrue: [ | eventBuffer type |
			eventBuffer := Array new: 8.
			MultiPointer getNextEvent: eventBuffer.
			type := eventBuffer at: 1.	
		
			type = EventTypeDeviceDiscovered ifTrue: [self handleDetectedEvent: eventBuffer].
			type = EventTypeDeviceRemoved ifTrue: [self handleRemovedEvent: eventBuffer].
			self moving ifTrue: [self sendHandMorphEvent: eventBuffer]].
	  ] repeat ]! !

!MultiPointerManager methodsFor: 'hand colors' stamp: 'mig 5/19/2009 16:28'!
releaseColor: aColor

	self colors at: aColor put: (self colors at: aColor) - 1! !

!MultiPointerManager methodsFor: 'hands' stamp: 'SR 8/18/2009 15:32'!
removeAllHands

	self handMorphs copy do: [:each | World removeHand: each].
! !

!MultiPointerManager methodsFor: 'event processing' stamp: 'SR 8/19/2009 14:13'!
sendHandMorphEvent: eventBuffer

	| type |
	type := eventBuffer at: 1.
	((type = EventTypeMouse) or: [type = EventTypeKeyboard]) ifFalse: [^ self].
! !

!MultiPointerManager methodsFor: 'system startup' stamp: 'SR 8/18/2009 15:46'!
shutdown

	self stopMoving.
	self stopQueueProcessor.
	MultiPointer unregisterSemaphores.! !

!MultiPointerManager methodsFor: 'processes' stamp: 'SR 8/18/2009 15:39'!
startMoving

	self moving: true.! !

!MultiPointerManager methodsFor: 'processes' stamp: 'SR 8/18/2009 15:40'!
startQueueProcessor

	self queueProcessor ifNotNil: [^self].
	
	self queueProcessor: 
		(self queueProcessorBlock
			forkAt: Processor userBackgroundPriority
			named: 'MultiPointer Queue Processor').! !

!MultiPointerManager methodsFor: 'system startup' stamp: 'SR 8/18/2009 15:46'!
startup

	MultiPointer registerSemaphores.
	self startQueueProcessor.! !

!MultiPointerManager methodsFor: 'processes' stamp: 'SR 8/18/2009 15:39'!
stopMoving

	self moving: false.! !

!MultiPointerManager methodsFor: 'processes' stamp: 'SR 8/18/2009 15:42'!
stopQueueProcessor

	self queueProcessor ifNotNil: [
		self queueProcessor terminate.
		self queueProcessor: nil.
		
		MultiPointer allPointers do: [:each |
			self destroyHandMorphFor: each].
		
		self handMorphs: OrderedCollection new].
! !

!MenuItemMorph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/22/2009 18:28'!
mouseDown: evt
	"Handle a mouse down event.
	Menu items get activated when the mouse is over them."

	"enable label editing" 
	evt shiftPressed ifTrue: [^ super mouseDown: evt].  

	"Bring menu to the front if it is obscured."
	self owner owner = self world ifTrue: [
		self owner comeToFront].

	(self isInDockingBar
			and:[isSelected])
		ifTrue:[
			evt hand newMouseFocus: nil.
			owner selectItem: nil event: evt. ]
		ifFalse:[
			"Redirect to menu for valid transitions"
			evt hand newMouseFocus: owner. 
			owner selectItem: self event: evt. ]
! !

!SharedQueue methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 5/4/2009 10:34'!
nextPut: value 
	"Send value through the receiver. If a Process has been suspended 
	waiting to receive a value through the receiver, allow it to proceed."

	value ifNil: [^ self].
	accessProtect
		critical: [writePosition > contentsArray size
						ifTrue: [self makeRoomAtEnd].
				 contentsArray at: writePosition put: value.
				 writePosition := writePosition + 1].
	readSynch signal.
	^value! !

!SharedQueue methodsFor: '*xpf-multipoint-core' stamp: 'SR 4/29/2009 17:24'!
nextPutAll: aCollection

	aCollection do: [:each | self nextPut: each].! !

!Morph methodsFor: '*xpf-multipoint-core-override' stamp: 'SR 6/22/2009 18:27'!
delete
	"Remove the receiver as a submorph of its owner and make its 
	new owner be nil."

	| aWorld |
	self removeHalo.
	aWorld := self world ifNil: [World].
	"Terminate genie recognition focus"
	"I encountered a case where the hand was nil, so I put in a little 
	protection - raa "
	" This happens when we are in an MVC project and open
	  a morphic window. - BG "
	aWorld ifNotNil: [
		aWorld handsDo:[:hand |
			self disableSubmorphFocusForHand: hand.
			hand
				releaseKeyboardFocus: self;
				releaseMouseFocus: self]].
	owner ifNotNil:[ self privateDelete.
		self player ifNotNilDo: [ :player |
			"Player must be notified"
			player noteDeletionOf: self fromWorld: aWorld]].! !
MultiPointerConstants initialize!
MultiPointerManager initialize!
